#ifdef DEBUG

#include "malloc.h"
#include "itox.h"
#include "strlen.h"
#include "null.h"

#define DEBUG_ADD
#else

typedef unsigned uintptr_t;
#define NULL 0L

#endif


#ifdef DEBUG
	#define print_str_debug print_str
	#define print_int_debug print_int
	#define assert(a) {if(!a) print_str("assert failure!"), cgc_terminate(254);}
#else
	#define print_str_debug(a) 
	#define print_int_debug(a) 
	#define assert(a)
#endif


#ifdef DEBUG
#define write_fd 2
void print_char(char c)
{
	write(write_fd,&c,1);
}
void print_str(char *s)
{
	write(write_fd,s,strlen(s));
}
void print_int(int x)
{
	char buf[100];
	itox(x,buf);
	write(write_fd,buf,strlen(buf));
}
#endif



typedef struct
{
        int edi;
        int esi;
        int ebp;
        int esp_dummy;
        int ebx;
        int edx;
        int ecx; 
        int eax;
        int flags;
} reg_values_t;

/*
global cgc_receive
cgc_receive:

orig_receive:
	mov    eax,0x3
	push   ebx
	push   ecx
	push   edx
	push   esi
	mov    ebx, [esp+0x14]
	mov    ecx, [esp+0x18]
	mov    edx, [esp+0x1c]
	mov    esi, [esp+0x20]

                         pushf
                         pusha
                         push retPC
	int    0x80 ==>  NOP callback to buffered_receive()


retPC:

                         popa
                         popf

	pop    esi
	pop    edx
	pop    ecx
	pop    ebx
	ret    
*/


#define STORAGE_SIZE 128
static char storage[STORAGE_SIZE];      
static int cursor_pos = 0;
static int bytes_buffered = 0;

int buffered_receive(unsigned int ret_pc, reg_values_t rv)
{
	// err = receive(fd,buf,count,rx_bytes)
 //       int fd = rv.ebx;
//	char *buf = (char*) rv.ecx;
//	int  count = rv.edx;
//	int  *rx_bytes = (int*) rv.esi;

        int bytes_out = 0;

        if (bytes_buffered == 0)
        {   
                cursor_pos = 0;

                // handle the case where nothing's buffered
                // and requested size > STORAGE_SIZE
                if (rv.edx > STORAGE_SIZE) 
                {   
                        // call out to original receive()
                        return rv.eax = cgc_receive(rv.ebx, (char*)rv.ecx, rv.edx, (int*) rv.esi);
                }   
        }   
        else
        {   
                // drain out buffer up to count bytes
                while (bytes_buffered > 0 && bytes_out < rv.edx) {
                        *((char*)rv.ecx) = storage[cursor_pos++];
			(char*)rv.ecx++;
			bytes_out++;
			bytes_buffered--;
		}

                if (bytes_buffered == 0 || cursor_pos == STORAGE_SIZE) {
                        cursor_pos = 0; 
                }   

                // if we're lucky, we've copied out all the requested data
                rv.edx -= bytes_out;
                if (rv.edx == 0)
                {   
                        if ((int*)rv.esi)
                                *((int*)rv.esi) = bytes_out;
                        return rv.eax = 0;
                }   
        }   

        // fill up the storage area again
        int num_bytes;
        int err_code = cgc_receive(rv.ebx, &storage[cursor_pos], STORAGE_SIZE - cursor_pos, &num_bytes);
        if (err_code == 0)  
        {   
                int bytes_to_copy = rv.edx;
                if (num_bytes < rv.edx) {
                        bytes_to_copy = num_bytes;
                }   

                // copy out to user buffer
                while (bytes_to_copy--) {
			*((char*)rv.ecx) = storage[cursor_pos++];
			(char*)rv.ecx++;
                }   
 
                bytes_buffered = num_bytes - rv.edx; 
                if (bytes_buffered >= 0)  
                { 
                        if ((int*)rv.esi)
                                *((int*)rv.esi) = rv.edx + bytes_out;
                }  
                else  
                { 
			// the user requested more bytes that we can hold in storage area... 
			// return only the bytes we have
                        // higher-level code has to be able to cope w/ the case 
                        // where the OS returns fewer bytes than requested anyhow 
                        if ((int*)rv.esi)
                                *((int*)rv.esi) = num_bytes + bytes_out;
                        bytes_buffered = 0; 
                } 
        }  
 
        return rv.eax = err_code; 
}
